//
//  DTMFGenerator.m
//  DTMF
//
//  Created by Benjamin Ragheb on Sat Jun 07 2003.
//  Copyright (c) 2003 __MyCompanyName__. All rights reserved.
//

#import "DTMFGenerator.h"

/*
 DTMF stands for Dual-Tone Multi-Frequency.  Each DTMF signal is the sum of two
 pure tones from the set of frequencies below:

                 1209Hz 1336Hz 1477Hz 1633Hz
                   |      |      |      |
                  .-.    .-.    .-.    .-.
          697Hz -( 1 )--( 2 )--( 3 )--( A )
                  >-<    >-<    >-<    >-<
          770Hz -( 4 )--( 5 )--( 6 )--( B )
                  >-<    >-<    >-<    >-<
          852Hz -( 7 )--( 8 )--( 9 )--( C )
                  >-<    >-<    >-<    >-<
          941Hz -( * )--( 0 )--( # )--( D )
                  `-'    `-'    `-'    `-'

 (Figure from http://www.ece.utexas.edu/~mason/codesign/dtmf.html)

 Signal duration: never below 23 ms; more than 40 ms is guaranteed to accept.
 Gap must be at least 10 ms.
 */

static NSCharacterSet *charSet697Hz = nil;
static NSCharacterSet *charSet770Hz;
static NSCharacterSet *charSet852Hz;
static NSCharacterSet *charSet941Hz;
static NSCharacterSet *charSet1209Hz;
static NSCharacterSet *charSet1336Hz;
static NSCharacterSet *charSet1477Hz;
//static NSCharacterSet *charSet1633Hz = nil; // "ABCD" buttons

#define DTMFSamplesPerSecond 8012

#define DTMFMinimumToneDuration 0.023
#define DTMFMinimumGapDuration 0.010

#define InitializeCharacterSet(name, string) \
name = [[NSCharacterSet characterSetWithCharactersInString:string] retain]

typedef struct {
    UInt32 magic;
    UInt32 headerSize;
    UInt32 dataSize;
    UInt32 encoding;
    UInt32 sampleRate;
    UInt32 numberOfChannels;
    UInt32 info;
} AUHeader;

@implementation DTMFGenerator

- (id)init
{
    self = [super init];
    if (self) {
        soundData = [[NSMutableData alloc] init];
        workBuffer = nil;
        workBufferLength = 0;
        defaultToneDuration = 0.100;
        defaultGapDuration = 0.050;
        defaultPauseDuration = 1.000;
        if (charSet697Hz == nil) {
            // Rows
            InitializeCharacterSet(charSet697Hz, @"123abcABCdefDEF");
            InitializeCharacterSet(charSet770Hz, @"456ghiGHIjklJKLmnoMNO");
            InitializeCharacterSet(charSet852Hz, @"789pqrsPQRStuvTUVwxyzWXYZ");
            InitializeCharacterSet(charSet941Hz, @"*0#");
            // Columns
            InitializeCharacterSet(charSet1209Hz, @"147*ghiGHIpqrsPQRS");
            InitializeCharacterSet(charSet1336Hz, @"2580abcABCjklJKLtuvTUV");
            InitializeCharacterSet(charSet1477Hz, @"369#defDEFmnoMNOwxyzWXYZ");
        }
    }
    return self;
}

- (void)dealloc
{
    [soundData release];
    if (workBuffer != nil) NSZoneFree([self zone], workBuffer);
    if (temporaryFilePath != nil) {
        [[NSFileManager defaultManager] removeFileAtPath:temporaryFilePath
                                                 handler:nil];
        [temporaryFilePath release];
    }
    [self dealloc];
}

- (void)_ensureWorkBufferSize:(unsigned int)size
{
    if (size > workBufferLength) {
        if (workBuffer != nil) NSZoneFree([self zone], workBuffer);
        workBufferLength = size;
        workBuffer = NSZoneMalloc([self zone], workBufferLength);
    }
}

- (int)_samplesForDuration:(double)duration
{
    return (int)(duration * DTMFSamplesPerSecond);
}

- (void)clear
{
    AUHeader *header;

    [soundData setLength:sizeof(AUHeader)];
    header = [soundData mutableBytes];
    header->magic = '.snd';
    header->headerSize = sizeof(AUHeader);
    header->dataSize = 0xFFFFFFFF;
    header->encoding = 2; // 8 bit Linear PCM
    header->sampleRate = DTMFSamplesPerSecond;
    header->numberOfChannels = 1;
    header->info = 'DTMF';
}

- (void)appendToneFrequency:(double)freq1 frequency:(double)freq2 duration:(double)duration
{
    int numberOfSamples = [self _samplesForDuration:duration];
    int n;
    double coeff1 = 2.0 * M_PI * freq1 / DTMFSamplesPerSecond;
    double coeff2 = 2.0 * M_PI * freq2 / DTMFSamplesPerSecond;
    double sample;

    [self _ensureWorkBufferSize:numberOfSamples];
    for (n = 0; n < numberOfSamples; n++) {
        sample = (63.0 * ( sin( n * coeff1 ) + sin( n * coeff2 ) ));
        workBuffer[n] = (SInt8)sample;
    }
    [soundData appendBytes:workBuffer length:numberOfSamples];
}

- (void)appendSilenceDuration:(double)duration
{
    [soundData increaseLengthBy:[self _samplesForDuration:duration]];
}

- (void)appendToneForCharacter:(unichar)x duration:(double)duration gap:(double)gap
{
    double freq1, freq2;

    if (x == ',') { [self appendSilenceDuration:defaultPauseDuration]; return; }
    if      ([charSet697Hz characterIsMember:x]) { freq1 = 697.0; }
    else if ([charSet770Hz characterIsMember:x]) { freq1 = 770.0; }
    else if ([charSet852Hz characterIsMember:x]) { freq1 = 852.0; }
    else if ([charSet941Hz characterIsMember:x]) { freq1 = 941.0; }
    else return;
    if      ([charSet1209Hz characterIsMember:x]) { freq2 = 1209.0; }
    else if ([charSet1336Hz characterIsMember:x]) { freq2 = 1336.0; }
    else if ([charSet1477Hz characterIsMember:x]) { freq2 = 1477.0; }
    //else if ([charSet1633Hz characterIsMember:x]) { freq2 = 1633.0; }
    else return;
    [self appendToneFrequency:freq1 frequency:freq2 duration:duration];
    [self appendSilenceDuration:gap];
}

- (void)appendToneForCharacter:(unichar)character
{
    [self appendToneForCharacter:character
                        duration:defaultToneDuration
                             gap:defaultGapDuration];
}

- (void)appendTonesForString:(NSString *)string
{
    int index;

    for (index = 0; index < [string length]; index++) {
        [self appendToneForCharacter:[string characterAtIndex:index]];
    }
}

- (double)toneDuration
{
    return defaultToneDuration;
}

- (void)setToneDuration:(double)seconds
{
    defaultToneDuration = seconds;
    if (seconds < DTMFMinimumToneDuration) {
        NSLog(@"Warning: Recommended minimum DTMF tone duration is %5.3f ms.",
              DTMFMinimumToneDuration);
    }
}

- (double)gapDuration
{
    return defaultGapDuration;
}

- (void)setGapDuration:(double)seconds
{
    defaultGapDuration = seconds;
    if (seconds < DTMFMinimumGapDuration) {
        NSLog(@"Warning: Recommended minimum DTMF gap duration is %5.3f ms.",
              DTMFMinimumGapDuration);
    }
}

- (double)pauseDuration
{
    return defaultPauseDuration;
}

- (void)setPauseDuration:(double)seconds
{
    defaultPauseDuration = seconds;
}

- (BOOL)writeToFile:(NSString *)path
{
    AUHeader *header = [soundData mutableBytes];

    header->dataSize = [soundData length];
    return [soundData writeToFile:path atomically:NO];
}

- (NSSound *)sound
{
    NSSound *snd;
    
    if (temporaryFilePath == nil) {
        NSString *path = NSTemporaryDirectory();
        path = [path stringByAppendingPathComponent:NSUserName()];
        path = [path stringByAppendingFormat:@"_DTMF_%x", self];
        path = [path stringByAppendingPathExtension:@"au"];
        temporaryFilePath = [path retain];
        NSLog(@"Saving audio file to %@", temporaryFilePath);
    }
    [self writeToFile:temporaryFilePath];
    snd = [[NSSound alloc] initWithContentsOfFile:temporaryFilePath
                                      byReference:NO];
    return [snd autorelease];
}   

@end
