//
//  DTMFSound.m
//  DTMF
//
//  Created by Benjamin Ragheb on Thu Jun 05 2003.
//  Copyright (c) 2003 __MyCompanyName__. All rights reserved.
//

/*
 .au file format info:
 http://astronomy.swin.edu.au/~pbourke/dataformats/au/

 DTMF generation info:
 http://www.hut.fi/~then/mytexts/dtmf_generation.html

               1209 Hz 1336 Hz 1477 Hz 1633 Hz

                        ABC     DEF
 697 Hz          1       2       3       A

                GHI     JKL     MNO
 770 Hz          4       5       6       B

                PRS     TUV     WXY
 852 Hz          7       8       9       C

                        oper
 941 Hz          *       0       #       D
 */

#import "DTMFSound.h"
#include <math.h>

#define DTMFSamplesPerSecond 8012
//#define DTMFSamplesPerSecond 44100

typedef struct {
    UInt32 signature;
    UInt32 headerSize;
    UInt32 dataSize;
    UInt32 encodingFormat;
    UInt32 sampleRate;
    UInt32 numberOfChannels;
} DTMFDataHeader;

static NSCharacterSet *charSet697Hz = nil;
static NSCharacterSet *charSet770Hz;
static NSCharacterSet *charSet852Hz;
static NSCharacterSet *charSet941Hz;
static NSCharacterSet *charSet1209Hz;
static NSCharacterSet *charSet1336Hz;
static NSCharacterSet *charSet1477Hz;
static NSCharacterSet *charSet1633Hz = nil; // "ABCD" buttons

#define DTMFInitializeCharacterSet(name, string) \
name = [[NSCharacterSet characterSetWithCharactersInString:string] retain]

void DTMFInitializeCharacterSets()
{
    if (charSet697Hz == nil) {
        // Rows
        DTMFInitializeCharacterSet(charSet697Hz, @"123abcABCdefDEF");
        DTMFInitializeCharacterSet(charSet770Hz, @"456ghiGHIjklJKLmnoMNO");
        DTMFInitializeCharacterSet(charSet852Hz, @"789pqrsPQRStuvTUVwxyzWXYZ");
        DTMFInitializeCharacterSet(charSet941Hz, @"*0#");
        // Columns
        DTMFInitializeCharacterSet(charSet1209Hz, @"147*ghiGHIpqrsPQRS");
        DTMFInitializeCharacterSet(charSet1336Hz, @"2580abcABCjklJKLtuvTUV");
        DTMFInitializeCharacterSet(charSet1477Hz, @"369#defDEFmnoMNOwxyzWXYZ");
    }
}

NSMutableData *DTMFCreateData(double durationCapacity)
{
    int numberOfSamples = ((double)DTMFSamplesPerSecond * durationCapacity);
    DTMFDataHeader header;
    NSMutableData *soundData;
    
    header.signature = '.snd';
    header.headerSize = 24;
    header.dataSize = -1;
    header.encodingFormat = 2; /* 8-bit linear PCM */
    //header.encodingFormat = 3; /* 16-bit linear PCM */
    header.sampleRate = DTMFSamplesPerSecond;
    header.numberOfChannels = 1;

    soundData = [[NSMutableData alloc] initWithCapacity:(numberOfSamples + 24)];
    [soundData appendBytes:&header length:sizeof(header)];
    return [soundData autorelease];
}

// method should use increaseLengthBy: and mutableBytes for efficiency
void DTMFDataAppendTone(NSMutableData *soundData, double frequency1,
                        double frequency2, double duration)
{
    int numberOfSamples = (int)((double)DTMFSamplesPerSecond * duration);
    int n;
    SInt8 s;
    //SInt16 s2;
    float coeff1 = 2.0 * M_PI * frequency1 / DTMFSamplesPerSecond;
    float coeff2 = 2.0 * M_PI * frequency2 / DTMFSamplesPerSecond;

    NSLog(@"Appending Tone: %e Hz and %e Hz for %e seconds",
          frequency1, frequency2, duration);
    for (n = 0; n < numberOfSamples; n++) {
        s = 63 * ( sin( n * coeff1 ) + sin( n * coeff2 ) );
        [soundData appendBytes:&s length:1];
        //s2 = 16383 * ( sin( n * coeff1 ) + sin( n * coeff2 ) );
        //[soundData appendBytes:&s2 length:2];
    }
}

void DTMFDataAppendSilence(NSMutableData *soundData, float duration)
{
    int numberOfSamples = (int)((float)DTMFSamplesPerSecond * duration);
    NSLog(@"Appending Silence: %e seconds", duration);
    [soundData increaseLengthBy:numberOfSamples];
}


NSData *DTMFDataFromDialingString(NSString *dialingString, double toneDuration)
{
    NSMutableData *soundData;
    int index;
    unichar x;
    double freq1, freq2;

    DTMFInitializeCharacterSets();
    soundData = DTMFCreateData(toneDuration * [dialingString length]);
    for (index = 0; index < [dialingString length]; index++) {
        x = [dialingString characterAtIndex:index];
        if (x == ',') { DTMFDataAppendSilence(soundData, 1.0); continue; }
        if ([charSet697Hz characterIsMember:x]) { freq1 = 697.0; }
        else if ([charSet770Hz characterIsMember:x]) { freq1 = 770.0; }
        else if ([charSet852Hz characterIsMember:x]) { freq1 = 852.0; }
        else if ([charSet941Hz characterIsMember:x]) { freq1 = 941.0; }
        else continue;
        if ([charSet1209Hz characterIsMember:x]) { freq2 = 1209.0; }
        else if ([charSet1336Hz characterIsMember:x]) { freq2 = 1336.0; }
        else if ([charSet1477Hz characterIsMember:x]) { freq2 = 1477.0; }
        else if ([charSet1633Hz characterIsMember:x]) { freq2 = 1633.0; }
        else continue;
        DTMFDataAppendTone(soundData, freq1, freq2, toneDuration);
    }
    return soundData;
}
