//
//  DTMFSound.m
//  DTMF
//
//  Created by Benjamin Ragheb on Thu Jun 05 2003.
//  Copyright (c) 2003 __MyCompanyName__. All rights reserved.
//

/*
 AIFF file format info:
 http://astronomy.swin.edu.au/~pbourke/dataformats/aiff/
 
 DTMF generation info:
 http://www.hut.fi/~then/mytexts/dtmf_generation.html

               1209 Hz 1336 Hz 1477 Hz 1633 Hz

                        ABC     DEF
 697 Hz          1       2       3       A

                GHI     JKL     MNO
 770 Hz          4       5       6       B

                PRS     TUV     WXY
 852 Hz          7       8       9       C

                        oper
 941 Hz          *       0       #       D


 AIFF
 16 bit Linear PCM
 Mono
 2.205 kHz
 */

#import "DTMFSound.h"
#include <math.h>

//#define DTMFSamplesPerSecond 22050
#define DTMFSamplesPerSecond 8012

static NSCharacterSet *charSet697Hz = nil;
static NSCharacterSet *charSet770Hz;
static NSCharacterSet *charSet852Hz;
static NSCharacterSet *charSet941Hz;
static NSCharacterSet *charSet1209Hz;
static NSCharacterSet *charSet1336Hz;
static NSCharacterSet *charSet1477Hz;
static NSCharacterSet *charSet1633Hz = nil; // "ABCD" buttons

#define DTMFInitializeCharacterSet(name, string) \
name = [[NSCharacterSet characterSetWithCharactersInString:string] retain]

void DTMFInitializeCharacterSets()
{
    if (charSet697Hz == nil) {
        // Rows
        DTMFInitializeCharacterSet(charSet697Hz, @"123abcABCdefDEF");
        DTMFInitializeCharacterSet(charSet770Hz, @"456ghiGHIjklJKLmnoMNO");
        DTMFInitializeCharacterSet(charSet852Hz, @"789pqrsPQRStuvTUVwxyzWXYZ");
        DTMFInitializeCharacterSet(charSet941Hz, @"*0#");
        // Columns
        DTMFInitializeCharacterSet(charSet1209Hz, @"147*ghiGHIpqrsPQRS");
        DTMFInitializeCharacterSet(charSet1336Hz, @"2580abcABCjklJKLtuvTUV");
        DTMFInitializeCharacterSet(charSet1477Hz, @"369#defDEFmnoMNOwxyzWXYZ");
    }
}

NSMutableData *DTMFCreateData(double durationCapacity)
{
    int numberOfSamples = ((double)DTMFSamplesPerSecond * durationCapacity);
    return [NSMutableData dataWithCapacity:numberOfSamples];
}

// method should use increaseLengthBy: and mutableBytes for efficiency
void DTMFDataAppendTone(NSMutableData *soundData, double frequency1,
                        double frequency2, double duration)
{
    int numberOfSamples = (int)((double)DTMFSamplesPerSecond * duration);
    int n;
    SInt8 s;
    //SInt16 s;
    float coeff1 = 2.0 * M_PI * frequency1 / DTMFSamplesPerSecond;
    float coeff2 = 2.0 * M_PI * frequency2 / DTMFSamplesPerSecond;

    //NSLog(@"Appending Tone: %e Hz and %e Hz for %e seconds",
    //      frequency1, frequency2, duration);
    for (n = 0; n < numberOfSamples; n++) {
        s = (SInt8)(63.0 * ( sin( n * coeff1 ) + sin( n * coeff2 ) ));
        [soundData appendBytes:&s length:1];
        //s = (SInt16)(16383.0 * ( sin( n * coeff1 ) + sin( n * coeff2 ) ));
        //[soundData appendBytes:&s length:2];
    }
}

void DTMFDataAppendSilence(NSMutableData *soundData, float duration)
{
    int numberOfSamples = (int)((float)DTMFSamplesPerSecond * duration);
    //NSLog(@"Appending Silence: %e seconds", duration);
    [soundData increaseLengthBy:numberOfSamples];
}


NSData *DTMFDataFromDialingString(NSString *dialingString, double toneDuration)
{
    NSMutableData *soundData;
    int index;
    unichar x;
    double freq1, freq2;

    DTMFInitializeCharacterSets();
    soundData = DTMFCreateData(toneDuration * [dialingString length]);
    for (index = 0; index < [dialingString length]; index++) {
        x = [dialingString characterAtIndex:index];
        if (x == ',') { DTMFDataAppendSilence(soundData, 1.0); continue; }
        if ([charSet697Hz characterIsMember:x]) { freq1 = 697.0; }
        else if ([charSet770Hz characterIsMember:x]) { freq1 = 770.0; }
        else if ([charSet852Hz characterIsMember:x]) { freq1 = 852.0; }
        else if ([charSet941Hz characterIsMember:x]) { freq1 = 941.0; }
        else continue;
        if ([charSet1209Hz characterIsMember:x]) { freq2 = 1209.0; }
        else if ([charSet1336Hz characterIsMember:x]) { freq2 = 1336.0; }
        else if ([charSet1477Hz characterIsMember:x]) { freq2 = 1477.0; }
        else if ([charSet1633Hz characterIsMember:x]) { freq2 = 1633.0; }
        else continue;
        DTMFDataAppendTone(soundData, freq1, freq2, toneDuration);
        DTMFDataAppendSilence(soundData, 0.05);
    }
    return soundData;
}

typedef struct {
    UInt32 formChunkID;   // 'FORM'
    UInt32 formChunkSize;
    UInt32 formChunkType; // 'AIFF'

    UInt32 commChunkID;   // 'COMM'
    UInt32 commChunkSize; // 18
    UInt16 numberOfChannels; // 1 (monophonic)
    UInt32 numberOfSampleFrames;
    UInt16 sampleSize; // 16 (bits)
    Float80 sampleRate; // 22050.0 (Hz)
    
    UInt32 ssndChunkID;   // 'SSND'
    UInt32 ssndChunkSize; // length + 8
    UInt32 ssndChunkOffset; // 0
    UInt32 ssndChunkBlockSize; // 0
} AIFFHeader;

NSData *AIFFDataFromSoundData(NSData *soundData)
{
    NSMutableData *aiffData;
    AIFFHeader header;

    header.formChunkID = 'FORM';
    header.formChunkSize = 4 + (8 + 18) + (8 + 8 + [soundData length]);
    header.formChunkType = 'AIFF';

    header.commChunkID = 'COMM';
    header.commChunkSize = 18;
    header.numberOfChannels = 1;
    header.numberOfSampleFrames = [soundData length] / 2;
    header.sampleSize = 16;
    header.sampleRate.exp = 0x400D;       // 0 100000000001110 
    header.sampleRate.man[0] = 0xAC44;    // 101011000100010 0
    header.sampleRate.man[1] = 0x0000;    // 0000000000000000
    header.sampleRate.man[2] = 0x0000;    // 0000000000000000
    header.sampleRate.man[3] = 0x0000;    // 0000000000000000
    
    header.ssndChunkID = 'SSND';
    header.ssndChunkSize = 8 + [soundData length];
    header.ssndChunkOffset = 0;
    header.ssndChunkBlockSize = 0;

    aiffData = [NSMutableData dataWithCapacity:([soundData length] +
                                                sizeof(header))];
    [aiffData appendBytes:&header length:sizeof(header)];
    [aiffData appendData:soundData];
    return aiffData;
}

typedef struct {
    UInt32 magic;
    UInt32 headerSize;
    UInt32 dataSize;
    UInt32 encoding;
    UInt32 sampleRate;
    UInt32 numberOfChannels;
    UInt32 info;
} AUHeader;

#define SND_FORMAT_LINEAR_8             (2)
#define SND_FORMAT_LINEAR_16            (3)
#define SND_RATE_CODEC          (8012.8210513)
#define SND_RATE_LOW            (22050.0)
#define SND_RATE_HIGH           (44100.0)
#define SND_RATE_LOW_PC         (11025.0)

NSData *AUDataFromSoundData(NSData *soundData)
{
    NSMutableData *auData;
    AUHeader header;

    header.magic = '.snd';
    header.headerSize = sizeof(AUHeader);
    header.dataSize = [soundData length];
    header.encoding = SND_FORMAT_LINEAR_8; // 16 bit linear PCM
    header.sampleRate = DTMFSamplesPerSecond;
    header.numberOfChannels = 1;
    header.info = 0;
    auData = [NSMutableData dataWithCapacity:([soundData length] +
                                              sizeof(header))];
    [auData appendBytes:&header length:sizeof(header)];
    [auData appendData:soundData];
    return auData;
}
