/*
 	DTMFGenerator.h
 
 	Created by Benjamin Ragheb on Sat Jun 07 2003.
 	Copyright (c) 2003 Benjamin Ragheb.  All rights reserved.
 */

#import <Cocoa/Cocoa.h>

#define kDTMFMinimumToneDuration 0.023
#define kDTMFMinimumGapDuration 0.010

#define kDTMFDefaultToneDuration 0.100
#define kDTMFDefaultGapDuration 0.050
#define kDTMFDefaultPauseDuration 1.000

/*
 The sounds used for touch tone dialing are called Dual Tone Multiple Frequencies (DTMF) tones.  The DTMFGenerator class can generate these tones for use in your application.  Each instance of the class encapsulates an audio buffer to which you can append tones.  At any time you can use the "writeToFile:atomically:" to write an AIFF audio file to disk or the "sound" method to obtain an NSSound object which can be played on the computer's speakers.  If the tones are played into a telephone mouthpiece, the phone will behave as if the number were dialed.
 */

@interface DTMFGenerator : NSObject
{
    NSMutableData *soundData;
    SInt8 *workBuffer;
    unsigned int workBufferLength;
    double defaultToneDuration;
    double defaultGapDuration;
    double defaultPauseDuration;
}

- (void)clear;
    // Discards all information in the audio buffer.  Call this method if you want to start generating a new set of tones.

- (void)appendToneFrequency:(double)freq1 frequency:(double)freq2 duration:(double)duration;
    // Appends a tone which is the sum of the given pair of frequencies.

- (void)appendSilenceDuration:(double)duration;
    // Appends silence to the audio buffer.

- (void)appendToneForCharacter:(unichar)x duration:(double)duration gap:(double)gap;
    // Appends a tone for the given character x, lasting "duration" seconds, followed by "gap" seconds of silence.  If x does not correspond to a DTMF tone, this method does nothing.

- (void)appendToneForCharacter:(unichar)character;
    // Appends a tone for the given character, lasting the default duration.

- (void)appendTonesForString:(NSString *)string;
    // Appends a sequence of tones and pauses based on the characters in string.  Alphanumeric characters will generate tones

- (double)toneDuration;
    // Returns the default tone duration.  This value is used by the methods [appendToneForCharacter:] and [appendTonesForString:].

- (void)setToneDuration:(double)seconds;
    // Sets the default tone duration.  If the value is below the recommended minimum, a warning is displayed on the console.  This value is used by the methods [appendToneForCharacter:] and [appendTonesForString:].

- (double)gapDuration;
    // Returns the default gap duration.

- (void)setGapDuration:(double)seconds;
    // Sets the default duration of the gap between tones.  If the value is below the recommended minimum, a warning is displayed on the console.

- (double)pauseDuration;
    // Returns the default pause duration.

- (void)setPauseDuration:(double)seconds;
    // Sets the default pause duration.

- (NSData *)aiffData;
    // Returns a copy of the audio data in AIFF format.

- (NSSound *)sound;
    // Returns an NSSound object initialized with the contents of the audio buffer.

@end
