//
//  DTMFGenerator.m
//  DTMF
//
//  Created by Benjamin Ragheb on Sat Jun 07 2003.
//  Copyright (c) 2008 Benjamin Ragheb. All rights reserved.
//

#import "DTMFGenerator.h"

/*
 DTMF stands for Dual-Tone Multi-Frequency.  Each DTMF signal is the sum of two
 pure tones from the set of frequencies below:

                      1209 Hz 1336 Hz 1477 Hz 1633 Hz
        
                               ABC     DEF
        697 Hz          1       2       3       A
        
                       GHI     JKL     MNO
        770 Hz          4       5       6       B
        
                       PRS     TUV     WXY
        852 Hz          7       8       9       C
        
                              oper
        941 Hz          *       0       #       D
 

 Signal duration: never below 23 ms; more than 40 ms is guaranteed to accept.
 Gap between tones must be at least 10 ms.
  */

static NSCharacterSet *charSet697Hz = nil;
static NSCharacterSet *charSet770Hz;
static NSCharacterSet *charSet852Hz;
static NSCharacterSet *charSet941Hz;
static NSCharacterSet *charSet1209Hz;
static NSCharacterSet *charSet1336Hz;
static NSCharacterSet *charSet1477Hz;
//static NSCharacterSet *charSet1633Hz = nil; // "ABCD" buttons

// If you change the sample rate, be sure to adjust the AIFF header code below.
#define DTMFSamplesPerSecond 8012

#define InitializeCharacterSet(name, string) \
name = [[NSCharacterSet characterSetWithCharactersInString:string] retain]

/*
 AIFFHeader defines the header fields for a minimal AIFF audio file.  This is 
 the only format that NSSound initWithData: will accept.  (It will accept other
 formats from disk, however.)  It is necessary to tell the compiler to use
 mac68k alignment, otherwise it will align the fields to 32 bits and generate
 a corrupt AIFF file.
 
 Audio Interchange File Format
 http://www.borg.com/~jglatt/tech/aiff.htm
 */
#pragma options align=mac68k

#define kAIFFExtraChunkSize 80

struct AIFFHeader {
    // BEGIN FORM CHUNK
    UInt32 formChunkID; // = 'FORM'
    SInt32 formChunkSize;
    UInt32 formTypeID; // = 'AIFF'
    // BEGIN AUTH CHUNK
    UInt32 extraChunkID;
    UInt32 extraChunkSize;
    UInt8 extraChunkText[kAIFFExtraChunkSize];
    // END AUTH CHUNK
    // BEGIN COMM CHUNK
    UInt32 commChunkID; // = 'COMM'
    UInt32 commChunkSize; // = 18
    SInt16 numberOfChannels; // = 1 (monophonic)
    UInt32 numberOfSampleFrames; 
    SInt16 bitsPerSamplePoint;
    UInt8 sampleFramesPerSecond[10]; // extended80
    // END COMM CHUNK
    // BEGIN SSND CHUNK
    UInt32 ssndChunkID; // = 'SSND'
    UInt32 ssndChunkSize; // = 8 + numberOfSampleFrames * int(bitsPerSamplePoint/8)
    UInt32 soundOffset; // = 0
    UInt32 blockSize; // = 0
    // END SSND CHUNK
    // END FORM CHUNK
};

typedef struct AIFFHeader AIFFHeader;


/*
 The sample rate (AIFFHeader.sampleFramesPerSecond) is stored in 80-bit extended
 floating point format.  As far as I can tell, GCC does not deal with this 
 natively and I'm not sure how to encode numbers by hand.  The following numbers
 were taken from some other AIFF files.
 
  8000 Hz = 400B FA00 0000 0000 0000
  8012 Hz = 400B FA60 0000 0000 0000
 
 11127 Hz = 400C ADDD 1748 0000 0000
 22254 Hz = 400D ADDC 0000 0000 0000
 
 11025 Hz = 400C AC44 0000 0000 0000
 22050 Hz = 400D AC44 0000 0000 0000
 44100 Hz = 400E AC44 0000 0000 0000
 88200 Hz = 400F AC44 0000 0000 0000
 */
static const UInt8 kDefaultSampleFramesPerSecond[] = {
    0x40, 0x0B, 0xFA, 0x60, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};



#pragma options align=reset


@implementation DTMFGenerator

/*
 Initialize the generator object.  Sets default values and initializes the 
 global character sets if necessary.
 */
- (id)init
{
    self = [super init];
    if (self) {
        soundData = [[NSMutableData alloc] init];
        workBuffer = nil;
        workBufferLength = 0;
        defaultToneDuration = kDTMFDefaultToneDuration;
        defaultGapDuration = kDTMFDefaultGapDuration;
        defaultPauseDuration = kDTMFDefaultPauseDuration;
        if (charSet697Hz == nil) {
            // Rows
            InitializeCharacterSet(charSet697Hz, @"123abcABCdefDEF");
            InitializeCharacterSet(charSet770Hz, @"456ghiGHIjklJKLmnoMNO");
            InitializeCharacterSet(charSet852Hz, @"789pqrsPQRStuvTUVwxyzWXYZ");
            InitializeCharacterSet(charSet941Hz, @"*0#");
            // Columns
            InitializeCharacterSet(charSet1209Hz, @"147*ghiGHIpqrsPQRS");
            InitializeCharacterSet(charSet1336Hz, @"2580abcABCjklJKLtuvTUV");
            InitializeCharacterSet(charSet1477Hz, @"369#defDEFmnoMNOwxyzWXYZ");
        }
    }
    return self;
}


/*
 Releases the generator object.
 */
- (void)dealloc
{
    [soundData release];
    if (workBuffer != nil) NSZoneFree([self zone], workBuffer);
    [super dealloc];
}


/*
 Ensures that the sound data buffer is at least a specified size.
 */
- (void)_ensureWorkBufferSize:(unsigned int)size
{
    if (size > workBufferLength) {
        if (workBuffer != nil) NSZoneFree([self zone], workBuffer);
        workBufferLength = size;
        workBuffer = NSZoneMalloc([self zone], workBufferLength);
    }
}


- (int)_samplesForDuration:(double)duration
{
    return (int)(duration * DTMFSamplesPerSecond);
}


- (void)clear
{
    [soundData setLength:sizeof(AIFFHeader)];
}


- (void)appendToneFrequency:(double)freq1 frequency:(double)freq2 duration:(double)duration
{
    int numberOfSamples = [self _samplesForDuration:duration];
    int n;
    double coeff1 = 2.0 * M_PI * freq1 / DTMFSamplesPerSecond;
    double coeff2 = 2.0 * M_PI * freq2 / DTMFSamplesPerSecond;
    double sample;

    [self _ensureWorkBufferSize:numberOfSamples];
    for (n = 0; n < numberOfSamples; n++) {
        sample = (63.0 * ( sin( n * coeff1 ) + sin( n * coeff2 ) ));
        workBuffer[n] = (SInt8)sample;
    }
    [soundData appendBytes:workBuffer length:numberOfSamples];
}


- (void)appendSilenceDuration:(double)duration
{
    [soundData increaseLengthBy:[self _samplesForDuration:duration]];
}


- (void)appendToneForCharacter:(unichar)x duration:(double)duration gap:(double)gap
{
    double freq1, freq2;

    if (x == ',') {
        [self appendSilenceDuration:defaultPauseDuration];
        return;
    }
    // select first frequency
    if      ([charSet697Hz characterIsMember:x]) { freq1 = 697.0; }
    else if ([charSet770Hz characterIsMember:x]) { freq1 = 770.0; }
    else if ([charSet852Hz characterIsMember:x]) { freq1 = 852.0; }
    else if ([charSet941Hz characterIsMember:x]) { freq1 = 941.0; }
    else return;
    // select second frequency
    if      ([charSet1209Hz characterIsMember:x]) { freq2 = 1209.0; }
    else if ([charSet1336Hz characterIsMember:x]) { freq2 = 1336.0; }
    else if ([charSet1477Hz characterIsMember:x]) { freq2 = 1477.0; }
    //else if ([charSet1633Hz characterIsMember:x]) { freq2 = 1633.0; }
    else return;
    // append tone to buffer
    [self appendToneFrequency:freq1 frequency:freq2 duration:duration];
    [self appendSilenceDuration:gap];
}


- (void)appendToneForCharacter:(unichar)character
{
    [self appendToneForCharacter:character
                        duration:defaultToneDuration
                             gap:defaultGapDuration];
}


- (void)appendTonesForString:(NSString *)string
{
    int index;

    for (index = 0; index < [string length]; index++) {
        [self appendToneForCharacter:[string characterAtIndex:index]];
    }
}


- (double)toneDuration
{
    return defaultToneDuration;
}


- (void)setToneDuration:(double)seconds
{
    defaultToneDuration = seconds;
    if (seconds < kDTMFMinimumToneDuration) {
        NSLog(@"Warning: DTMF tone duration set to %5.3f; recommended minimum is %5.3f ms.",
              seconds, kDTMFMinimumToneDuration);
    }
}


- (double)gapDuration
{
    return defaultGapDuration;
}


- (void)setGapDuration:(double)seconds
{
    defaultGapDuration = seconds;
    if (seconds < kDTMFMinimumGapDuration) {
        NSLog(@"Warning: DTMF gap duration set to %5.3f; recommended minimum is %5.3f ms.",
              seconds, kDTMFMinimumGapDuration);
    }
}


- (double)pauseDuration
{
    return defaultPauseDuration;
}


- (void)setPauseDuration:(double)seconds
{
    defaultPauseDuration = seconds;
}


- (void)_setupAIFFHeader
{
    AIFFHeader *header = [soundData mutableBytes];
    
    header->formChunkID = NSSwapHostIntToBig('FORM');
    header->formChunkSize = NSSwapHostIntToBig([soundData length] - 8);
    header->formTypeID = NSSwapHostIntToBig('AIFF');
    header->extraChunkID = NSSwapHostIntToBig('ANNO');
    header->extraChunkSize = NSSwapHostIntToBig(kAIFFExtraChunkSize);
    strcpy((char *)header->extraChunkText, "Generated by SoundDial");
    header->commChunkID = NSSwapHostIntToBig('COMM');
    header->commChunkSize = NSSwapHostIntToBig(18);
    header->numberOfChannels = NSSwapHostShortToBig(1);
    header->numberOfSampleFrames = NSSwapHostIntToBig([soundData length] - sizeof(AIFFHeader));
    header->bitsPerSamplePoint = NSSwapHostShortToBig(8);
    memcpy(header->sampleFramesPerSecond, kDefaultSampleFramesPerSecond, 10);
    header->ssndChunkID = NSSwapHostIntToBig('SSND');
    header->ssndChunkSize = NSSwapHostIntToBig([soundData length] - sizeof(AIFFHeader) + 8);
    header->soundOffset = 0;
    header->blockSize = 0;
}


- (NSData *)aiffData
{
    [self _setupAIFFHeader];
    return [[soundData copy] autorelease];
}


- (NSSound *)sound
{
    [self _setupAIFFHeader];
    return [[[NSSound alloc] initWithData:soundData] autorelease];
}   

@end
