#import "SoundDialApp.h"
#import "DTMFGenerator.h"
#import "DTMFGeneratorSaveFile.h"

static NSString *kToneDurationMillisDefaultsKey = @"ToneDuration";
static NSString *kGapDurationMillisDefaultsKey = @"GapDuration";
static NSString *kPauseDurationMillisDefaultsKey = @"PauseDuration";
static NSString *kDialStringDefaultsKey = @"DialString";
static NSString *kDialStringHistoryDefaultsKey = @"DialStringHistory";

static NSString *kABPlugInsDirectoryName = @"Address Book Plug-Ins";
static NSString *kABPlugInBundleName = @"SoundDial AB Plug-In";
static NSString *kABPlugInBundleExtension = @"bundle";

@implementation SoundDialApp

+ (void)initialize
{
	[[NSUserDefaults standardUserDefaults] registerDefaults:[NSDictionary dictionaryWithObjectsAndKeys:
		[NSNumber numberWithInt:(kDTMFDefaultToneDuration * 1000)], kToneDurationMillisDefaultsKey, 
		[NSNumber numberWithInt:(kDTMFDefaultGapDuration * 1000)], kGapDurationMillisDefaultsKey, 
		[NSNumber numberWithInt:(kDTMFDefaultPauseDuration * 1000)], kPauseDurationMillisDefaultsKey,
		@"", kDialStringDefaultsKey,
		[NSArray array], kDialStringHistoryDefaultsKey,
		nil]];
}

- (id)init
{
	if ([super init]) {
		dtmfGenerator = [[DTMFGenerator alloc] init];
		dialStringHistory = [[[NSUserDefaults standardUserDefaults] arrayForKey:kDialStringHistoryDefaultsKey] mutableCopy];
	}
	return self;
}

- (void)awakeFromNib
{
    [self reloadUserDefaults];
    [dialField setStringValue:[[NSUserDefaults standardUserDefaults] stringForKey:kDialStringDefaultsKey]];
}

- (void)reloadUserDefaults
{
    NSUserDefaults *userDefaults = [NSUserDefaults standardUserDefaults];
    double durationMillis;

    durationMillis = [[userDefaults objectForKey:kToneDurationMillisDefaultsKey] doubleValue];
    [dtmfGenerator setToneDuration:(durationMillis / 1000.0)];
    durationMillis = [[userDefaults objectForKey:kGapDurationMillisDefaultsKey] doubleValue];
    [dtmfGenerator setGapDuration:(durationMillis / 1000.0)];
    durationMillis = [[userDefaults objectForKey:kPauseDurationMillisDefaultsKey] doubleValue];
    [dtmfGenerator setPauseDuration:(durationMillis / 1000.0)];
}

- (void)dealloc
{
	[dialStringHistory release];
    [dtmfGenerator release];
    [super dealloc];
}

- (void)updateDialStringHistory 
{
	NSString *dialString = [dialField stringValue];
	unsigned historyIndex = [dialStringHistory indexOfObject:dialString];
	if (historyIndex != 0) {
		if (historyIndex != NSNotFound) {
			[dialStringHistory removeObjectAtIndex:historyIndex];
		}
		[dialStringHistory insertObject:dialString atIndex:0];
		if ([dialStringHistory count] > 5) {
			[dialStringHistory removeLastObject];
		}
		[dialField reloadData];
	}
}

- (IBAction)playTones:(id)sender
{
    if ([sound isPlaying]) {
        [sound stop];
    } else {
		[self updateDialStringHistory];
        [self reloadUserDefaults];
        [dtmfGenerator clear];
        [dtmfGenerator appendTonesForString:[dialField stringValue]];
        sound = [dtmfGenerator sound];
        [sound setDelegate:self];
        if ([sound play]) {
            [sound retain];
            [dialField setEnabled:NO];
            [dialButton setTitle:NSLocalizedString(@"Stop", nil)];
            [dialButton setAction:@selector(stopPlayingTones:)];
        } else {
            NSLog(@"Can't play the sound.");
        }
    }
}

- (IBAction)stopPlayingTones:(id)sender
{
    [sound stop];
}

- (BOOL)validateMenuItem:(NSMenuItem *)menuItem
{
    if ([menuItem action] == @selector(playTones:)) {
        return ![sound isPlaying];
    }
    if ([menuItem action] == @selector(stopPlayingTones:)) {
        return [sound isPlaying];
    }
    return YES;
}

- (IBAction)saveTones:(id)sender
{
    [self reloadUserDefaults];
    [dtmfGenerator clear];
    [dtmfGenerator appendTonesForString:[dialField stringValue]];
    [dtmfGenerator beginSaveSheetWithFile:[dialField stringValue]
                           modalForWindow:mainWindow];
}

- (IBAction)restoreDefaults:(id)sender
{
    NSUserDefaults *userDefaults = [NSUserDefaults standardUserDefaults];
    
    [userDefaults setInteger:(kDTMFDefaultToneDuration * 1000) forKey:kToneDurationMillisDefaultsKey];
    [userDefaults setInteger:(kDTMFDefaultGapDuration * 1000) forKey:kGapDurationMillisDefaultsKey];
    [userDefaults setInteger:(kDTMFDefaultPauseDuration * 1000) forKey:kPauseDurationMillisDefaultsKey];
}

- (void)showErrorMessage:(NSString *)message informativeText:(NSString *)informativeText
{
	NSAlert *alert = [[NSAlert alloc] init];
	[alert addButtonWithTitle:@"OK"];
	[alert setMessageText:message];
	[alert setInformativeText:informativeText];
	[alert setAlertStyle:NSCriticalAlertStyle];
	[alert runModal];
	[alert release];
}

- (void)showInfoMessage:(NSString *)message informativeText:(NSString *)informativeText
{
	NSAlert *alert = [[NSAlert alloc] init];
	[alert addButtonWithTitle:@"OK"];
	[alert setMessageText:message];
	[alert setInformativeText:informativeText];
	[alert setAlertStyle:NSInformationalAlertStyle];
	[alert runModal];
	[alert release];
}

- (IBAction)installABPlugIn:(id)sender
{
	NSAlert *alert = [[NSAlert alloc] init];
	[alert addButtonWithTitle:@"Current User"];
	[alert addButtonWithTitle:@"All Users"];
	[alert addButtonWithTitle:@"Cancel"];
	[alert setMessageText:@"Where to install?"];
	[alert setInformativeText:@"Where should the Address Book Plug-In be installed?"];
	[alert setAlertStyle:NSInformationalAlertStyle];
	
	int userChoice = [alert runModal];
	[alert release];
	
	NSSearchPathDomainMask domainMask;
	switch (userChoice) {
		case NSAlertFirstButtonReturn:
			domainMask = NSUserDomainMask;
			break;
		case NSAlertSecondButtonReturn:
			domainMask = NSLocalDomainMask;
			break;
		default:
			return;
	}			

	NSArray *array = NSSearchPathForDirectoriesInDomains(NSLibraryDirectory, domainMask, YES);
	if ([array count] < 1) {
		[self showErrorMessage:@"No Library Directories" 
			   informativeText:@"No Library directories found."];
		return;
	}

	NSString *dstDirPath = [[array objectAtIndex:0] stringByAppendingPathComponent:kABPlugInsDirectoryName];
	
    NSString *srcPath = [[NSBundle mainBundle] pathForResource:kABPlugInBundleName
														ofType:kABPlugInBundleExtension];
	
	NSFileManager *fileMgr = [NSFileManager defaultManager];
	
	BOOL dstIsDir;
	BOOL dstExists = [fileMgr fileExistsAtPath:dstDirPath isDirectory:&dstIsDir];
	
	if (dstExists && !dstIsDir) {
		[self showErrorMessage:@"File in the way" 
			   informativeText:@"The plug-in cannot be installed because a file exists where the Address Book Plug-Ins directory ought to be."];
		return;
	} else if (! dstExists) {
		BOOL createSucceeded = [fileMgr createDirectoryAtPath:dstDirPath attributes:nil];
		if (! createSucceeded) {
			[self showErrorMessage:@"Cannot create plug-ins directory" 
				   informativeText:@"The plug-in cannot be installed because the Address Book Plug-Ins directory does not exist and cannot be created."];
			return;
		}
	}
	
	NSString *dstPath = [dstDirPath stringByAppendingPathComponent:[srcPath lastPathComponent]];
	
	if ([fileMgr fileExistsAtPath:dstPath]) {
		BOOL removeSucceeded = [fileMgr removeFileAtPath:dstPath handler:nil];
		if (! removeSucceeded) {
			[self showErrorMessage:@"Plug-in already installed" 
				   informativeText:@"The plug-in cannot be installed because it appears to already be installed and cannot be removed."];
			return;
		}
	}
	
	BOOL copySucceeded = [fileMgr copyPath:srcPath toPath:dstPath handler:nil];
	
	if (copySucceeded) {
		[self showInfoMessage:@"Plug-in installed successfully" 
			  informativeText:@"The plug-in was successfully installed.  You must quit and relaunch Address Book for changes to take effect."];
	} else {		
		[self showErrorMessage:@"Installation failed" 
			   informativeText:@"The plug-in was not installed because of a failure while copying files."];
	}
}

- (IBAction)removeABPlugIn:(id)sender
{
	BOOL nothingFound = YES;
	NSFileManager *fileMgr = [NSFileManager defaultManager];
	NSArray *array = NSSearchPathForDirectoriesInDomains(NSLibraryDirectory, NSUserDomainMask | NSLocalDomainMask, YES);
	int i;
	for (i = 0; i < [array count]; i++) {
		NSString *pluginsPath = [[array objectAtIndex:i] stringByAppendingPathComponent:kABPlugInsDirectoryName];
		NSString *bundlePath = [pluginsPath stringByAppendingPathComponent:[kABPlugInBundleName stringByAppendingPathExtension:kABPlugInBundleExtension]];
		if ([fileMgr fileExistsAtPath:bundlePath]) {
			nothingFound = NO;
			BOOL removeSucceeded = [fileMgr removeFileAtPath:bundlePath handler:nil];
			if (removeSucceeded) {
				[self showInfoMessage:@"Plug-in removed successfully" 
					  informativeText:[NSString stringWithFormat:@"The plug-in was successfully removed from %@.", pluginsPath]];
			} else {
				[self showErrorMessage:@"Cannot remove plug-in" 
					   informativeText:[NSString stringWithFormat:@"The plug-in could not be removed from %@.", pluginsPath]];
			}
		}
	}
	if (nothingFound) {
		[self showErrorMessage:@"No plug-ins found" 
			   informativeText:@"SoundDial did not find any copies of the plug-in to remove."];
	}
}

/*
 The phone number is finished playing, so we can re-enable the text field and
 button in the main window.
 */
- (void)sound:(NSSound *)aSound didFinishPlaying:(BOOL)aBool
{
    [sound release];
    sound = nil;
    [dialField setEnabled:YES];
    [dialButton setTitle:NSLocalizedString(@"Play Tones", nil)];
    [dialButton setAction:@selector(playTones:)];
    [[NSApp mainWindow] makeFirstResponder:dialField];
}

/*
 To export a service
 */
- (void)playTones:(NSPasteboard *)pboard userData:(NSString *)userData error:(NSString **)error
{
    NSString *pboardString;
    
    if (![[pboard types] containsObject:NSStringPboardType]) {
        *error = NSLocalizedString(@"Error: couldn't encrypt text.",
                                   @"pboard couldn't give string.");
        return;
    }
    pboardString = [pboard stringForType:NSStringPboardType];
    if (!pboardString) {
        *error = NSLocalizedString(@"Error: couldn't encrypt text.",
                                   @"pboard couldn't give string.");
        return;
    }
    [dialField setStringValue:pboardString];
    [self playTones:nil];
}

/*
 Register ourself as a service provider.
 */
- (void)applicationDidFinishLaunching:(NSNotification *)notification
{
    [NSApp setServicesProvider:self];
    NSUpdateDynamicServices();
}

/*
 If the user clicks on the dock icon and no windows are visible, this method
 will be called.  We'll bring the main window to the front.
 */
- (BOOL)applicationOpenUntitledFile:(NSApplication *)theApplication
{
    [mainWindow makeKeyAndOrderFront:self];
    return YES;
}

/*
 The user might be switching to the Address Book, so we synchronize the 
 defaults database so the Address Book Plug-In will load the correct settings.
 */
- (void)applicationWillResignActive:(NSNotification *)aNotification
{
    [[NSUserDefaults standardUserDefaults] synchronize];
}

/*
 Save the contents of the telephone number field so that it will be there
 the next time the application is launched.
 */
- (void)applicationWillTerminate:(NSNotification *)aNotification
{
    NSUserDefaults *userDefaults = [NSUserDefaults standardUserDefaults];
    
    [userDefaults setObject:[dialField stringValue] forKey:kDialStringDefaultsKey];
	[userDefaults setObject:dialStringHistory forKey:kDialStringHistoryDefaultsKey];
}

// NSComboBoxDataSource

- (int)numberOfItemsInComboBox:(NSComboBox *)historyBox
{
	return [dialStringHistory count];
}

- (id)comboBox:(NSComboBox *)historyBox objectValueForItemAtIndex:(int)index
{
	return [dialStringHistory objectAtIndex:index];
}

@end
